import io
from itertools import groupby
from pathlib import Path

import pytest
from selenium import webdriver

import ee.digikey as dk

basedir = Path(__file__).parent
static_copies = basedir / "digikey" / "static-copies"  # type: Path

digikey = dk.Digikey()
client = dk.DigikeyClient(digikey, cache_dir=basedir / "digikey" / "http_cache", on_download=print)
driver: webdriver.Chrome = None
force_refresh = False  # Set to True to always fetch the updated html files


@pytest.mark.digikey
def test_digikey_1(tmpdir):
    content = cache_file("https://www.digikey.com/products/en?lang=en&site=us&pageSize=10", "TCR2LF18LM(CTTR-ND")

    res = client.parse_string(content)
    assert res.response_type == dk.SearchResponseTypes.SINGLE
    p = res.products[0]
    assert p.part_number == "TCR2LF18LM(CTTR-ND"
    assert p.mpn == "TCR2LF18,LM(CT"
    assert len(p.attributes) > 5

    repo = dk.DigikeyRepository(digikey, str(tmpdir))
    x = io.StringIO()
    p.to_ini(repo._make_configparser()).write(x)
    print(x.getvalue())
    repo.save(p)
    repo = dk.DigikeyRepository(digikey, str(tmpdir))
    repo.load_all()
    ps = repo.find_by_mpn(p.mpn)
    assert ps
    y = io.StringIO()
    p.to_ini(repo._make_configparser()).write(y)
    assert x.getvalue() == y.getvalue()


@pytest.mark.digikey
def test_digikey_2():
    content = cache_file("https://www.digikey.com/products/en?lang=en&site=us&&pageSize=500", "TCR2LF")

    res = client.parse_string(content)
    assert res.response_type == dk.SearchResponseTypes.MANY
    [print("dpn={}, mpn={}".format(p.part_number, p.mpn)) for p in res.products]
    assert len(res.products) > 10

    for mpn, parts in groupby(sorted(res.products, key=lambda p: p.mpn), lambda p: p.mpn):
        parts = list(parts)
        print("mpn={}, parts={}".format(mpn, [p.part_number for p in parts]))

        dpn = parts[0].part_number
        print("Downloading {} as {}".format(mpn, dpn))

        content = cache_file("https://www.digikey.com/products/en?lang=en&site=us&pageSize=500", dpn)

        res = client.parse_string(content)

        assert res.response_type == dk.SearchResponseTypes.SINGLE
        p = res.products[0]
        assert p.part_number == dpn
        assert p.mpn == mpn


@pytest.mark.digikey
def test_digikey_3():
    content = cache_file("https://www.digikey.com/products/en?lang=en&site=us&pageSize=10", "RS1MTR")

    res = client.parse_string(content)
    assert res.response_type == dk.SearchResponseTypes.MANY
    [print("dpn={}, mpn={}".format(p.part_number, p.mpn)) for p in res.products]
    assert len(res.products) > 0

    p = next((p for p in res.products if p.part_number == "1655-1501-1-ND"), None)
    assert p.mpn == "RS1MTR"
    assert p.url == "/product-detail/en/smc-diode-solutions/RS1MTR/1655-1501-1-ND/6022946"


def cache_file(url, keyword):
    path = static_copies / "search-{}.html".format(keyword)

    if force_refresh and path.is_file():
        path.unlink()

    if not path.is_file():
        path.parent.mkdir(parents=True, exist_ok=True)
        from urllib.parse import quote
        url = url + "&keywords=" + quote(keyword)
        print("GET {}".format(url))

        global driver
        if driver is None:
            options = webdriver.ChromeOptions()
            driver = webdriver.Chrome(chrome_options=options)

        driver.get(url)

        with open(str(path), "w") as f:
            f.write(driver.page_source)
        assert path.stat().st_size > 0

    with open(str(path), "r") as f:
        content = f.read()

    return content


try:
    driver.close()
except Exception:
    pass
